@@ -110,6 +110,8 @@ pub struct GitConfig {
     pub link_parsers: Vec<LinkParser>,
     /// Exclude commits that are not matched by any commit parser.
     pub filter_commits: bool,
+    /// Fail on a commit that is not matched by any commit parser.
+    pub fail_on_unmatched_commit: bool,
     /// Regex to select git tags that represent releases.
     #[serde(with = "serde_regex", default)]
     pub tag_pattern: Option<Regex>,
@@ -187,6 +189,9 @@ pub struct RemoteConfig {
     /// Bitbucket remote.
     #[serde(default)]
     pub bitbucket: Remote,
+    /// Azure DevOps remote.
+    #[serde(default)]
+    pub azure_devops: Remote,
 }
 
 impl RemoteConfig {
@@ -208,6 +213,10 @@ impl RemoteConfig {
         if self.bitbucket.is_set() {
             return true;
         }
+        #[cfg(feature = "azure_devops")]
+        if self.azure_devops.is_set() {
+            return true;
+        }
         false
     }
 
@@ -229,6 +238,10 @@ impl RemoteConfig {
         {
             self.bitbucket.native_tls = Some(true);
         }
+        #[cfg(feature = "azure_devops")]
+        {
+            self.azure_devops.native_tls = Some(true);
+        }
     }
 }
 
@@ -354,10 +367,10 @@ impl Bump {
     /// This function also logs the returned value.
     pub fn get_initial_tag(&self) -> String {
         if let Some(tag) = self.initial_tag.clone() {
-            warn!("No releases found, using initial tag '{tag}' as the next version.");
+            log::warn!("No releases found, using initial tag '{tag}' as the next version");
             tag
         } else {
-            warn!("No releases found, using {DEFAULT_INITIAL_TAG} as the next version.");
+            log::warn!("No releases found, using {DEFAULT_INITIAL_TAG} as the next version");
             DEFAULT_INITIAL_TAG.into()
         }
     }
@@ -499,7 +512,7 @@ impl Config {
         .filter_map(|v| v.as_ref())
         {
             if supported_path.exists() {
-                debug!("Using configuration file from: {:?}", supported_path);
+                log::debug!("Using configuration file from: {:?}", supported_path);
                 return Some(supported_path.to_path_buf());
             }
         }
