@@ -1,5 +1,5 @@
 use std::io;
-use std::path::{Path, PathBuf};
+use std::path::{self, Path, PathBuf};
 use std::result::Result as StdResult;
 
 use git2::{
@@ -45,32 +45,73 @@ pub struct SubmoduleRange {
 }
 
 impl Repository {
-    /// Initializes (opens) the repository.
-    pub fn init(path: PathBuf) -> Result<Self> {
-        if path.exists() {
-            let inner = GitRepository::discover(&path).or_else(|err| {
-                let jujutsu_path = path.join(".jj").join("repo").join("store").join("git");
-                if jujutsu_path.exists() {
-                    GitRepository::open_bare(&jujutsu_path)
+    /// Opens a repository from the given path.
+    ///
+    /// If `search_parents` is true, it will traverse up through parent
+    /// directories to find a repository.
+    fn open(path: PathBuf, search_parents: bool) -> Result<Self> {
+        if !path.exists() {
+            return Err(Error::IoError(io::Error::new(
+                io::ErrorKind::NotFound,
+                format!("repository path not found: {}", path.display()),
+            )));
+        }
+
+        let inner = GitRepository::open(&path)
+            .or_else(|err| {
+                // Optionally search for a Jujutsu repository layout
+                let mut current = Some(path.as_path());
+                while let Some(dir) = current {
+                    let jujutsu_path = dir.join(".jj/repo/store/git");
+                    if jujutsu_path.exists() {
+                        return GitRepository::open_bare(&jujutsu_path);
+                    }
+                    // Only continue searching if enabled
+                    if !search_parents {
+                        break;
+                    }
+                    current = dir.parent();
+                }
+                Err(err)
+            })
+            // If still not found, try discover if traversal is enabled
+            .or_else(|err| {
+                if search_parents {
+                    GitRepository::discover(&path)
                 } else {
                     Err(err)
                 }
             })?;
-            let changed_files_cache_path = inner
-                .path()
-                .join(env!("CARGO_PKG_NAME"))
-                .join(CHANGED_FILES_CACHE);
-            Ok(Self {
-                inner,
-                path,
-                changed_files_cache_path,
-            })
-        } else {
-            Err(Error::IoError(io::Error::new(
-                io::ErrorKind::NotFound,
-                "repository path not found",
-            )))
-        }
+
+        let changed_files_cache_path = inner
+            .path()
+            .join(env!("CARGO_PKG_NAME"))
+            .join(CHANGED_FILES_CACHE);
+
+        Ok(Self {
+            inner,
+            path,
+            changed_files_cache_path,
+        })
+    }
+
+    /// Discover a repository from the given path by traversing up through
+    /// parent directories.
+    ///
+    /// It first looks for a Git repository using [`GitRepository::discover`].
+    /// If no Git repository is found, it checks for a Jujutsu repository layout
+    /// (`.jj/repo/store/git`) in this directory and its parents.
+    pub fn discover(path: PathBuf) -> Result<Self> {
+        Self::open(path, true)
+    }
+
+    /// Attempts to open an already-existing repository at the given path.
+    ///
+    /// It tries to open the repository as a normal or bare Git repository located
+    /// exactly at `path`. If that fails, it falls back to checking for a Jujutsu
+    /// repository layout (`.jj/repo/store/git`) **only in the specified directory**.
+    pub fn init(path: PathBuf) -> Result<Self> {
+        Self::open(path, false)
     }
 
     /// Returns the path of the repository.
@@ -189,17 +230,21 @@ impl Repository {
                 // submodule updated
                 Some(format!("{}..{}", old_file_id, new_file_id))
             };
-            trace!("Release commit range for submodules: {:?}", range);
+            log::trace!("Release commit range for submodules: {:?}", range);
             delta.new_file().path().and_then(Path::to_str).zip(range)
         });
         // iterate through all path diffs and find corresponding submodule if
         // possible
         let submodule_range = before_and_after_deltas.filter_map(|(path, range)| {
+            // NOTE:
+            // libgit2 recommends using `git_submodule_open`, whereas `git_repository_discover` is
+            // used here. Since it seems to be working fine for now, we don't think we
+            // should change this. Just leaving this message as a reminder.
             let repository = self
                 .inner
                 .find_submodule(path)
                 .ok()
-                .and_then(|submodule| Self::init(submodule.path().into()).ok());
+                .and_then(|submodule| Self::discover(submodule.path().into()).ok());
             repository.map(|repository| SubmoduleRange { repository, range })
         });
         Ok(submodule_range.collect())
@@ -210,10 +255,10 @@ impl Repository {
     /// It removes the leading `./` and adds `**` to the end if the pattern is a
     /// directory.
     fn normalize_pattern(pattern: Pattern) -> Pattern {
-        let star_added = match pattern.as_str().chars().last() {
-            Some('/' | '\\') => Pattern::new(&format!("{pattern}**"))
-                .expect("failed to add '**' to the end of glob"),
-            _ => pattern,
+        let star_added = if pattern.as_str().ends_with(path::MAIN_SEPARATOR) {
+            Pattern::new(&format!("{pattern}**")).expect("failed to add '**' to the end of glob")
+        } else {
+            pattern
         };
         match star_added.as_str().strip_prefix("./") {
             Some(stripped) => {
@@ -305,11 +350,11 @@ impl Repository {
                     cache_key,
                     v,
                 ) {
-                    error!("Failed to set cache for repo {:?}: {e}", self.path);
+                    log::error!("Failed to set cache for repo {:?}: {e}", self.path);
                 }
             }
             Err(e) => {
-                error!("Failed to serialize cache for repo {:?}: {e}", self.path);
+                log::error!("Failed to serialize cache for repo {:?}: {e}", self.path);
             }
         }
 
@@ -500,7 +545,7 @@ impl Repository {
                     .url()
                     .ok_or_else(|| Error::RepoError(String::from("failed to get the remote URL")))?
                     .to_string();
-                trace!("Upstream URL: {url}");
+                log::trace!("Upstream URL: {url}");
                 return find_remote(&url);
             }
         }
@@ -581,7 +626,7 @@ fn ssh_path_segments(url: &str) -> Result<Remote> {
 #[cfg(test)]
 mod test {
     use std::process::Command;
-    use std::{env, fs, str};
+    use std::{env, fs, io, str};
 
     use temp_dir::TempDir;
 
@@ -625,7 +670,7 @@ mod test {
     }
 
     fn get_repository() -> Result<Repository> {
-        Repository::init(
+        Repository::discover(
             PathBuf::from(env!("CARGO_MANIFEST_DIR"))
                 .parent()
                 .expect("parent directory not found")
@@ -779,7 +824,8 @@ mod test {
             .expect("failed to execute git init");
         assert!(output.status.success(), "git init failed {:?}", output);
 
-        let repo = Repository::init(temp_dir.path().to_path_buf()).expect("failed to init repo");
+        let repo =
+            Repository::discover(temp_dir.path().to_path_buf()).expect("failed to init repo");
         let output = Command::new("git")
             .args(["config", "user.email", "test@gmail.com"])
             .current_dir(temp_dir.path())
@@ -806,7 +852,21 @@ mod test {
     }
 
     #[test]
-    fn open_jujutsu_repo() {
+    fn repository_path_not_found() {
+        let path = PathBuf::from("/this/path/should/not/exist/123456789");
+        let result = Repository::discover(path.clone());
+        assert!(result.is_err());
+        match result {
+            Err(Error::IoError(err)) => {
+                assert_eq!(err.kind(), io::ErrorKind::NotFound);
+                assert!(err.to_string().contains("repository path not found"));
+            }
+            _ => panic!("expected IoError(NotFound)"),
+        }
+    }
+
+    #[test]
+    fn discover_jujutsu_repo() {
         let (repo, _temp_dir) = create_temp_repo();
         // working copy is the directory that contains the .git directory:
         let working_copy = repo.path;
@@ -822,6 +882,76 @@ mod test {
         fs::create_dir_all(&store).expect("failed to create dir");
         fs::rename(working_copy.join(".git"), store.join("git")).expect("failed to move git repo");
 
+        // Open repo from working copy, that contains the .jj directory
+        let repo = Repository::discover(working_copy).expect("failed to init repo");
+
+        // macOS canonical path for temp directories is in /private
+        // libgit2 forces the path to be canonical regardless of what we pass in
+        if repo.inner.path().starts_with("/private") {
+            assert_eq!(
+                repo.inner.path().strip_prefix("/private"),
+                store.join("git").strip_prefix("/"),
+                "open git repo in .jj/repo/store/"
+            );
+        } else {
+            assert_eq!(
+                repo.inner.path(),
+                store.join("git"),
+                "open git repo in .jj/repo/store/"
+            );
+        }
+    }
+
+    #[test]
+    fn propagate_error_if_no_repo_found() {
+        let temp_dir = TempDir::with_prefix("git-cliff-").expect("failed to create temp dir");
+
+        let path = temp_dir.path().to_path_buf();
+
+        let result = Repository::discover(path.clone());
+
+        assert!(result.is_err());
+        if let Err(error) = result {
+            assert!(
+                format!("{error:?}").contains(
+                    format!("could not find repository at '{}'", path.display()).as_str()
+                )
+            )
+        }
+    }
+
+    #[test]
+    fn repository_path_does_not_exist() {
+        let path = PathBuf::from("/this/path/should/not/exist/123456789");
+        let result = Repository::init(path.clone());
+        assert!(result.is_err());
+        match result {
+            Err(Error::IoError(err)) => {
+                assert_eq!(err.kind(), io::ErrorKind::NotFound);
+                assert!(err.to_string().contains("repository path not found"));
+            }
+            _ => panic!("expected IoError(NotFound)"),
+        }
+    }
+
+    #[test]
+    fn open_jujutsu_repo() {
+        let (repo, _temp_dir) = create_temp_repo();
+        // working copy is the directory that contains the .git directory:
+        let working_copy = repo.path;
+
+        // Make the Git repository bare and set HEAD
+        Command::new("git")
+            .args(["config", "core.bare", "true"])
+            .current_dir(&working_copy)
+            .status()
+            .expect("failed to make git repo non-bare");
+
+        // Move the Git repo into jj
+        let store = working_copy.join(".jj").join("repo").join("store");
+        fs::create_dir_all(&store).expect("failed to create dir");
+        fs::rename(working_copy.join(".git"), store.join("git")).expect("failed to move git repo");
+
         // Open repo from working copy, that contains the .jj directory
         let repo = Repository::init(working_copy).expect("failed to init repo");
 
@@ -843,7 +973,7 @@ mod test {
     }
 
     #[test]
-    fn propagate_error_if_no_repo_found() {
+    fn propagate_error_if_no_repo_exist() {
         let temp_dir = TempDir::with_prefix("git-cliff-").expect("failed to create temp dir");
 
         let path = temp_dir.path().to_path_buf();
